
from __future__ import annotations

import sys
from pathlib import Path


def is_relative_to(a: Path | str, b: Path | str) -> bool:
  """ Returns `True` if path *a* is relative to path *b*. A backfill for #Path.is_relative_to() for Python versions
  older than 3.9. """

  if sys.version_info < (3, 9):
    try:
      Path(a).relative_to(b)
    except ValueError:
      return False
    return True
  else:
    return Path(a).is_relative_to(b)




import collections
import functools
import typing as t

from nr.util.generic import T

T_OrderedSet = t.TypeVar('T_OrderedSet', bound='OrderedSet')


@functools.total_ordering
class OrderedSet(t.MutableSet[T]):

  def __init__(self, iterable: t.Optional[t.Iterable[T]] = None) -> None:
    self._index_map: t.Dict[T, int] = {}
    self._content: t.Deque[T] = collections.deque()
    if iterable is not None:
      self.update(iterable)

  def __repr__(self) -> str:
    if not self._content:
      return '%s()' % (type(self).__name__,)
    return '%s(%r)' % (type(self).__name__, list(self))

  def __iter__(self) -> t.Iterator[T]:
    return iter(self._content)

  def __reversed__(self) -> 'OrderedSet[T]':
    return OrderedSet(reversed(self._content))

  def __eq__(self, other: t.Any) -> bool:
    if isinstance(other, OrderedSet):
      return len(self) == len(other) and list(self) == list(other)
    return False

  def __le__(self, other: t.Any) -> bool:
      return all(e in other for e in self)

  def __len__(self) -> int:
    return len(self._content)

  def __contains__(self, key: t.Any) -> bool:
    return key in self._index_map

  def __getitem__(self, index: int) -> T:
    return self._content[index]

  def add(self, key: T) -> None:
    if key not in self._index_map:
      self._index_map[key] = len(self._content)
      self._content.append(key)

  def copy(self: T_OrderedSet) -> 'T_OrderedSet':
    return type(self)(self)

  def discard(self, key: T) -> None:
    if key in self._index_map:
      index = self._index_map.pop(key)
      del self._content[index]

  def pop(self, last: bool = True) -> T:
    if not self._content:
      raise KeyError('set is empty')
    key = self._content.pop() if last else self._content.popleft()
    self._index_map.pop(key)
    return key

  def update(self, iterable: t.Iterable[T]) -> None:
    for x in iterable:
      self.add(x)




from __future__ import annotations

import typing as t

from deprecated import deprecated

from nr.util.generic import T
from nr.util.iter import SequenceWalker as _SequenceWalker


@deprecated('use nr.util.iter.SequenceWalker instead')
class Scanner(_SequenceWalker[T]):

  @deprecated('use Scanner.safe_iter() instead')
  def ensure_advancing(self, strict_forward: bool = True) -> t.Iterator[T]:
    return self.safe_iter(strict_forward)




import typing_extensions as te


class ASGIApp(te.Protocol):

  def __call__(self, scope, receive, send):
    raise NotImplementedError


class WSGIApp(te.Protocol):

  def __call__(self, environ, start_response):
    raise NotImplementedError




from ._assure import assure
from ._coalesce import coalesce
from ._consumer import Consumer
from ._once import Once
from ._predicate import Predicate
from ._supplier import supplier_get, Supplier, ContextSupplier

__all__ = ['assure', 'coalesce', 'Consumer', 'Once', 'Predicate', 'supplier_get', 'Supplier', 'ContextSupplier']




import dataclasses
import os
import subprocess as sp
from . import AWSGILauncher, BaseAWSGIConfig


@dataclasses.dataclass
class GunicornRunner(AWSGILauncher, BaseAWSGIConfig):
  """
  Runs a WSGI application using [Gunicorn][0].

  [0]: https://gunicorn.org/
  """

  def launch(self, entrypoint: str, factory: bool) -> None:
    if factory:
      entrypoint += '()'
    command = ['gunicorn', entrypoint, '--bind', '{}:{}'.format(self.host, self.port)]
    if self.daemonize:
      command.append('--daemon')
    if self.files.pidfile:
      os.makedirs(os.path.dirname(self.files.pidfile), exist_ok=True)
      command += ['--pid', self.files.pidfile]
    if self.files.stdout:
      os.makedirs(os.path.dirname(self.files.stdout), exist_ok=True)
      command += ['--access-logfile', self.files.stdout]
    if self.files.stderr:
      os.makedirs(os.path.dirname(self.files.stderr), exist_ok=True)
      command += ['--error-logfile', self.files.stderr]
    if self.ssl:
      command += ['--certfile', self.ssl.cert, '--keyfile', self.ssl.key]
    if self.num_workers:
      command += ['--workers', str(self.num_workers)]
    command += self.additional_options
    sp.call(command)




import datetime
import io
import re
import typing as t
from dataclasses import dataclass
from .options import DatetimeComponentType, IFormatOption, FormatOptions

_T_datetime_format = t.TypeVar('_T_datetime_format', bound='_datetime_format')


@dataclass
class _datetime_format:
  format_str: str
  regex: 're.Pattern'
  seq: t.List[t.Union[str, IFormatOption]]

  @classmethod
  def compile(cls: t.Type[_T_datetime_format], format_str: str, regex_mode: bool = False) -> '_T_datetime_format':
    """
    Compiles a format string to a static regex representation for fast parsing.
    """

    combo_regex = io.StringIO()
    combo_regex.write('^')
    combo_sequence: t.List[t.Union[str, IFormatOption]] = []

    def append_seq(char: str) -> None:
      if combo_sequence and isinstance(combo_sequence[-1], str):
        combo_sequence[-1] += char
      else:
        combo_sequence.append(char)

    idx = 0
    while idx < len(format_str):
      if regex_mode and format_str[idx] in '().?':
        if format_str[idx] == '(':
          combo_regex.write('(?:')
        else:
          combo_regex.write(format_str[idx])
      elif regex_mode and format_str[idx] == '\\':
        # TODO(NiklasRosenstein): oob check
        idx += 1
        char = format_str[idx]
        combo_regex.write(re.escape(char))
        append_seq(char)
      elif format_str[idx] == '%' and len(format_str) > idx + 1:
        # TODO(NiklasRosenstein): Interpret %% as single %
        idx += 1
        char = format_str[idx]
        option = FormatOptions.get(char)
        if option is None:
          raise ValueError(f'unknown format option %{char}')
        combo_regex.write('(' + option.regex + ')')
        combo_sequence.append(option)
      else:
        char = format_str[idx]
        combo_regex.write(re.escape(char))
        append_seq(char)
      idx += 1

    combo_regex.write('$')
    return cls(format_str, re.compile(combo_regex.getvalue()), combo_sequence)

  def __repr__(self) -> str:
    return f'{type(self).__name__}({self.format_str!r})'


@dataclass
class date_format(_datetime_format):

  __repr__ = _datetime_format.__repr__

  def __post_init__(self) -> None:
    for item in self.seq:
      if isinstance(item, IFormatOption) and item.component.type != DatetimeComponentType.Date:
        raise ValueError(f'%{item.char} is an invalid format option for date_format')

  def parse_date(self, s: str) -> datetime.date:
    # TODO(NiklasRosenstein): Validate that the format string actually only captures date components.
    dt = datetime_format(self.format_str, self.regex, self.seq).parse_datetime(s)
    return dt.date()

  def format_date(self, d: datetime.date) -> str:
    # TODO(NiklasRosenstein): Validate that the format string actually only captures date components.
    dt = datetime.datetime(d.year, d.month, d.day)
    return datetime_format(self.format_str, self.regex, self.seq).format_datetime(dt)

  def to_datetime_format(self) -> 'datetime_format':
    return datetime_format(self.format_str, self.regex, self.seq)


@dataclass
class time_format(_datetime_format):

  __repr__ = _datetime_format.__repr__

  def __post_init__(self) -> None:
    for item in self.seq:
      if isinstance(item, IFormatOption) and item.component.type != DatetimeComponentType.Time:
        raise ValueError(f'%{item.char} is an invalid format option for time_format')

  def parse_time(self, s: str) -> datetime.time:
    # TODO(NiklasRosenstein): Validate that the format string actually only captures time components.
    dt = datetime_format(self.format_str, self.regex, self.seq).parse_datetime(s)
    return dt.time()

  def format_time(self, t: datetime.time) -> str:
    # TODO(NiklasRosenstein): Validate that the format string actually only captures time components.
    dt = datetime.datetime(1970, 1, 1, t.hour, t.minute, t.second, t.microsecond, t.tzinfo)
    return datetime_format(self.format_str, self.regex, self.seq).format_datetime(dt)

  def to_datetime_format(self) -> 'datetime_format':
    return datetime_format(self.format_str, self.regex, self.seq)


class datetime_format(_datetime_format):

  def parse_datetime(self, s: str) -> datetime.datetime:
    match = self.regex.match(s)
    if not match:
      raise ValueError(f'"{s}" does not match format "{self.format_str}"')
    kwargs = {'year': 1900, 'month': 1, 'day': 1, 'hour': 0}
    groups = iter(match.groups())
    for item in self.seq:
      if isinstance(item, IFormatOption):
        matched_string = next(groups)
        if matched_string is not None:
          kwargs[item.component.value] = item.parse_string(matched_string)
    return datetime.datetime(**kwargs)  # type: ignore

  def format_datetime(self, dt: datetime.datetime) -> str:
    result = io.StringIO()
    for item in self.seq:
      if isinstance(item, str):
        result.write(item)
      else:
        result.write(item.format_value(dt, getattr(dt, item.component.value)))
    return result.getvalue()




"""
Utility functions for regular expressions. Builds on top of the standard library #re module.
"""

import re
import typing as t


class MatchFullError(ValueError):
  """
  Raised when #match_full() cannot consume the full string.
  """

  def __init__(self, regex: 're.Pattern', string: str, endpos: int) -> None:
    self.regex = regex
    self.string = string
    self.endpos = endpos

  def __str__(self):
    return 'could not consume whole string with regex {} (got until position {})'.format(
      self.regex, self.endpos)


def match_full(expr: t.Union[str, 're.Pattern'], string: str) -> t.Iterable['re.Match']:
  """
  Matches *expr* from the start of *string* and expects that it can be matched throughout.
  If it fails to consume the full string, a #MatchAllError will be raised.
  """

  if isinstance(expr, str):
    expr = re.compile(expr)

  offset = 0
  while offset < len(string):
    match = expr.match(string, offset)
    if not match:
      raise MatchFullError(expr, string, offset)
    offset = match.end()
    yield match




from __future__ import annotations

import typing as t
from ._api import KeyValueStore

V = t.TypeVar('V')
V_Plain = t.TypeVar('V_Plain', bound=t.Union[int, float, str])
_Encoder = t.Callable[[V], bytes]
_Decoder = t.Callable[[bytes], V]


def _default_encoder(v: V) -> bytes:
  return str(v).encode('utf8')


def _default_decoder(t: t.Type[V]) -> _Decoder:
  def _decoder(v: bytes) -> V:
    return t(v.decode('utf8'))  # type: ignore
  return _decoder


class MappingAdapter(t.MutableMapping[str, V]):
  """
  Adapter for interacting with a #KeyValueStore as a #Mapping. The default constructor supports plain datatypes
  as values to represent an encoder/decoder.
  """

  def __init__(self, kv: KeyValueStore, decoder: t.Union[_Decoder, t.Type[V]], encoder: t.Optional[_Encoder] = None) -> None:
    self._kv = kv
    self._encoder = encoder or _default_encoder
    if isinstance(decoder, type):
      example_value = decoder()
      decoder = _default_decoder(decoder)
      decoded_value = decoder(self._encoder(example_value))
      assert decoded_value == example_value and type(decoded_value) == type(example_value), 'decoder/encoder mismatch'
    self._decoder = decoder

  def __getitem__(self, key: str) -> V:
    return self._decoder(self._kv.get(key))

  def __setitem__(self, key: str, value: V) -> None:
    return self._kv.set(key, self._encoder(value))

  def __delitem__(self, key: str) -> None:
    self._kv.delete(key)

  def __iter__(self) -> t.Iterator[str]:
    yield from self._kv.keys()

  def __len__(self) -> int:
    return self._kv.count()




from ._attribute import Attribute
from ._color import Color, SgrColor, SgrColorName, LutColor, TrueColor
from ._style import Style, StyleManager

__all__ = [
  'Attribute',
  'Color',
  'SgrColor',
  'SgrColorName',
  'LutColor',
  'TrueColor',
  'Style',
  'StyleManager',
]




""" Utilities for filesystem operations. """

from . import _chmod as chmod
from ._atomic import atomic_swap, atomic_write
from ._discovery import get_file_in_directory
from ._walk import recurse_directory, walk_up
from ._path import is_relative_to

__all__ = [
  'chmod',
  'atomic_swap',
  'atomic_write',
  'get_file_in_directory',
  'recurse_directory',
  'walk_up',
  'is_relative_to',
]




__all__ = [
  'tzlocal',
  'duration',
  'date_format',
  'datetime_format',
  'time_format',
  'parse_date',
  'parse_datetime',
  'parse_time',
  'format_date',
  'format_datetime',
  'format_date',
  'format_set',
  'JAVA_OFFSET_DATETIME',
  'ISO_8601',
]

import datetime
import time
import typing as t
from .duration import duration
from .format import date_format, datetime_format, time_format
from .format_sets import format_set, JAVA_OFFSET_DATETIME, ISO_8601


def tzlocal() -> datetime.tzinfo:
  offset = time.altzone if time.daylight else time.timezone
  return datetime.timezone(datetime.timedelta(seconds=-offset))


def parse_date(fmt: t.Union[format_set, date_format, str], s: str) -> datetime.date:
  if isinstance(fmt, str):
    fmt = date_format.compile(fmt)
  return fmt.parse_date(s)


def parse_datetime(fmt: t.Union[format_set, datetime_format, str], s: str) -> datetime.datetime:
  if isinstance(fmt, str):
    fmt = datetime_format.compile(fmt)
  return fmt.parse_datetime(s)


def parse_time(fmt: t.Union[format_set, time_format, str], s: str) -> datetime.time:
  if isinstance(fmt, str):
    fmt = time_format.compile(fmt)
  return fmt.parse_time(s)


def format_date(fmt: t.Union[format_set, date_format, str], d: datetime.date) -> str:
  if isinstance(fmt, str):
    fmt = date_format.compile(fmt)
  return fmt.format_date(d)


def format_datetime(fmt: t.Union[format_set, datetime_format, str], dt: datetime.datetime) -> str:
  if isinstance(fmt, str):
    fmt = datetime_format.compile(fmt)
  return fmt.format_datetime(dt)


def format_time(fmt: t.Union[format_set, time_format, str], t: datetime.time) -> str:
  if isinstance(fmt, str):
    fmt = time_format.compile(fmt)
  return fmt.format_time(t)




import typing as t
from nr.util.digraph import DiGraph, E, K, N
from nr.util.types import Comparable


def topological_sort(graph: DiGraph[K, N, E], sorting_key: t.Optional[t.Callable[[K], Comparable]] = None) -> t.Iterator[K]:
  """ Calculate the topological order for elements in the *graph*.

  @raises RuntimeError: If there is a cycle in the graph. """

  seen: set[K] = set()
  roots = graph.roots

  while roots:
    if seen & roots:
      raise RuntimeError(f'encountered a cycle in the graph at {seen & roots}')
    seen.update(roots)
    yield from roots
    roots = {
      k: None
      for n in roots for k in sorted(graph.successors(n), key=sorting_key)  # type: ignore
      if not graph.predecessors(k) - seen
    }.keys()

  if len(seen) != len(graph.nodes):
    raise RuntimeError(f'encountered a cycle in the graph (unreached nodes {set(graph.nodes) - seen})')




from __future__ import annotations

import typing as t
import typing_extensions as te

from nr.util.generic import T_co, U


def supplier_get(value: Supplier[T_co] | U) -> T_co | U:
  """ A helper method to invoke a supplier, if *value* is a supplier, or return the value directory. This is
  useful in particular for patterns where a static value as well as a supplier is a valid value for the variable.
  Note that the type *U* must not be callable, otherwise it is recognized as a supplier.

  __Example__

    site_name: Supplier[str | None] | str | None = ...
    value = supplier_get(site_name)
  """

  if callable(value):
    return value()
  return value


class Supplier(te.Protocol[T_co]):

  def __call__(self) -> T_co:
    ...


class ContextSupplier(Supplier[t.ContextManager[T_co]]):
  ...




""" General purpose utility library. """

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = "0.8.4"


# Backwards compatibility
from .functional._coalesce import coalesce
from .chaindict import ChainDict
from .optional import Optional
from .orderedset import OrderedSet
from .refreshable import Refreshable
from .stream import Stream




import enum
from ._utils import process_exists, process_terminate


class Pidfile:

  class Status(enum.Enum):
    STOPPED = 0
    RUNNING = 1
    UNKNOWN = 2

  def __init__(self, filename: str) -> None:
    self.name = filename

  def set_pid(self, pid: int) -> None:
    """
    Writes the *pid* into the file.
    """

    with open(self.name) as fp:
      fp.write(str(pid))

  def get_pid(self) -> int:
    """
    Returns the PID saved in the file. May raise a #FileNotFounError if the file does not
    exist or a #ValueError if the file does not contain an integer.
    """

    with open(self.name) as fp:
      return int(fp.readline().strip())

  def get_status(self) -> Status:
    """
    Determines the #Status of a process.
    """

    try:
      pid = self.get_pid()
    except FileNotFoundError:
      return Pidfile.Status.STOPPED
    except ValueError:
      return Pidfile.Status.UNKNOWN
    if process_exists(pid):
      return Pidfile.Status.RUNNING
    return Pidfile.Status.STOPPED

  def stop(self) -> None:
    """
    Stops the process referenced by the PID file. If the file does not exist or the
    process is not alive, this is a no-op.
    """

    try:
      pid = self.get_pid()
    except (FileNotFoundError, ValueError):
      return
    process_terminate(pid)




import abc
import datetime
import enum
import re
import typing as t
from dataclasses import dataclass


class DatetimeComponentType(enum.Enum):
  Date = enum.auto()
  Time = enum.auto()


class DatetimeComponent(enum.Enum):
  Year = 'year'
  Month = 'month'
  Day = 'day'
  Hour = 'hour'
  Minute = 'minute'
  Second = 'second'
  Microsecond = 'microsecond'
  Timezone = 'tzinfo'

  @property
  def type(self) -> DatetimeComponentType:
    if self in (DatetimeComponent.Year, DatetimeComponent.Month, DatetimeComponent.Day):
      return DatetimeComponentType.Date
    if self in (DatetimeComponent.Hour, DatetimeComponent.Minute, DatetimeComponent.Second, DatetimeComponent.Microsecond, DatetimeComponent.Timezone):
      return DatetimeComponentType.Time
    raise RuntimeError(f'unexpected Component enumeration value: {self!r}')


@dataclass
class _FormatOption:
  """
  Represents a special character in a date format string, which in turn represents a particular
  date or time component.
  """

  #: The character that the format option is represented in a format string. Must be a string of
  #: length 1 (one).
  char: str

  #: The component of a datetime construct that the format option represents.
  component: DatetimeComponent

  #: The regular expression that is used to capture the value represented by the format option
  #: when parsing a date/time string. This expression must not contain any capturing groups.
  regex: str


class IFormatOption(_FormatOption, metaclass=abc.ABCMeta):
  """
  Abstract metaclass on top of #_FormatOption, adding methods to the interface for parsing and
  formatting values represented by the format option.
  """

  @abc.abstractmethod
  def parse_string(self, s: str) -> t.Any:
    """ Called to parse then value extracted by the *regex*. """

  @abc.abstractmethod
  def format_value(self, dt: datetime.datetime, v: t.Any) -> str:
    """ Called to format the value of the respective #Component to a string. """


@dataclass
class NumericFormatOption(IFormatOption):
  """
  Default implementation for format options that represent a plain number component in the date.
  """

  format: t.Callable[[int], str]
  post_parse: t.Optional[t.Callable[[str], int]] = None

  def parse_string(self, s: str) -> t.Any:
    if self.post_parse is not None:
      return self.post_parse(s)
    return int(s)

  def format_value(self, dt: datetime.datetime, v: t.Any) -> str:
    assert isinstance(v, int), f'expected int to {self!r}.format_value()'
    return self.format(v)  # type: ignore


@dataclass
class TimezoneFormatOption(IFormatOption):

  regex: str = r'(?:Z|[-+]\d{2}(?::?\d{2})?)'

  def parse_string(self, s: str) -> datetime.tzinfo:
    match = re.match(self.regex, s)
    if not match:
      raise ValueError('not a timezone string: {!r}'.format(s))
    if s == 'Z':
      return datetime.timezone.utc
    else:
      s = s.replace(':', '')
      sign = -1 if s[0] == '-' else 1
      hours = int(s[1:3])
      minutes = int(s[3:5] or '00')
      seconds = sign * (hours * 3600 + minutes * 60)
      return datetime.timezone(datetime.timedelta(seconds=seconds))

  def format_value(self, dt: datetime.datetime, v: t.Any) -> str:
    assert v is None or isinstance(v, datetime.tzinfo), f'expected datetime.tzinfo, got {v!r}'
    if v is None:
      return ''
    elif v == datetime.timezone.utc:
      return 'Z'
    else:
      assert isinstance(v, datetime.tzinfo)
      utcoffset = v.utcoffset(dt)
      # NOTE Copied from CPython 3.7 datetime.py _format_offset()
      string = ''
      if utcoffset is not None:
        if utcoffset.days < 0:
          sign = "-"
          utcoffset = -utcoffset
        else:
          sign = "+"
        off = utcoffset.total_seconds()
        hh, mm = divmod(off, 60 * 60)
        mm, ss = divmod(mm, 60)
        ss, ms = divmod(ss, 1)
        string += "%s%02d:%02d" % (sign, hh, mm)
        if ss or ms:
          string += ":%02d" % ss
          if ms:
            string += '.%06d' % ms
      return string


class FormatOptions(enum.Enum):
  """
  Enumeration of all the available format options.
  """

  Year = NumericFormatOption('Y', DatetimeComponent.Year, r'\d{4}', lambda v: str(v).rjust(4, '0'))
  Month = NumericFormatOption('m', DatetimeComponent.Month, r'\d{2}', lambda v: str(v).rjust(2, '0'))
  Day = NumericFormatOption('d', DatetimeComponent.Day, r'\d{2}', lambda v: str(v).rjust(2, '0'))
  Hour = NumericFormatOption('H', DatetimeComponent.Hour, r'\d{2}', lambda v: str(v).rjust(2, '0'))
  Minute = NumericFormatOption('M', DatetimeComponent.Minute, r'\d{2}', lambda v: str(v).rjust(2, '0'))
  Second = NumericFormatOption('S', DatetimeComponent.Second, r'\d{2}', lambda v: str(v).rjust(2, '0'))
  Microsecond = NumericFormatOption('f', DatetimeComponent.Microsecond, r'\d+',
      format=lambda v: str(v).rjust(6, '0').rstrip('0') or '0',
      post_parse=lambda v: int(str(int(v) * (10 ** max(6-len(v), 0)))[:6]))
  Timezone = TimezoneFormatOption('z', DatetimeComponent.Timezone)

  @classmethod
  def get(cls, char: str) -> t.Optional[IFormatOption]:
    # TODO(NiklasRosenstein): Room for optimization by caching available format options in dict.
    return next((x.value for x in cls if x.value.char == char), None)




import abc
import dataclasses
import enum
import re


class Color(abc.ABC):

  @abc.abstractmethod
  def as_foreground(self) -> str: ...

  @abc.abstractmethod
  def as_background(self) -> str: ...


class SgrColorName(enum.Enum):
  BLACK = 0
  RED = 1
  GREEN = 2
  YELLOW = 3
  BLUE = 4
  MAGENTA = 5
  CYAN = 6
  WHITE = 7
  GRAY = 8
  DEFAULT = 9


@dataclasses.dataclass
class SgrColor(Color):
  """ Represents a color from the SGR space (see #SgrColorName). """

  name: SgrColorName
  bright: bool

  def __init__(self, name: SgrColorName, bright: bool = False) -> None:
    if isinstance(name, str):
      name = SgrColorName[name.upper()]
    self.name = name
    self.bright = bright

  def as_foreground(self) -> str:
    return str((90 if self.bright else 30) + self.name.value)

  def as_background(self) -> str:
    return str((100 if self.bright else 40) + self.name.value)


@dataclasses.dataclass
class LutColor(Color):
  """ Represents a LUT color, which is one of 216 colors. """

  index: int

  def as_foreground(self) -> str:
    return '38;5;' + str(self.index)

  def as_background(self) -> str:
    return '48;5;' + str(self.index)

  @classmethod
  def from_rgb(cls, r: int, g: int, b: int) -> 'LutColor':
    """
    Given RGB values in the range of [0..5], returns a #LutColor pointing
    to the color index that resembles the specified color coordinates.
    """

    def _check_range(name, value):
      if not (0 <= value < 6):
        raise ValueError('bad value for parameter "{}": {} ∉ [0..5]'.format(name, value))

    _check_range('r', r)
    _check_range('g', g)
    _check_range('b', b)

    return cls((16 + 36 * r) + (6 * g) + b)


class TrueColor(Color):
  """ Represents a true color comprised of three color components. """

  r: int
  g: int
  b: int

  def as_foreground(self) -> str:
    return '38;2;{};{};{}'.format(self.r, self.g, self.b)

  def as_background(self) -> str:
    return '48;2;{};{};{}'.format(self.r, self.g, self.b)


def parse_color(color_string: str) -> Color:
  """ Parses a color string of one of the following formats and returns a corresponding #SgrColor, #LutColor or
  #TrueColor.

  * `<color_name>`, `BRIGHT_<color_name>`: #SgrColor (case insensitive, underline optional)
  * `%rgb`, `$xxx`: #LutColor
  * `#cef`, `#cceeff`: #TrueColor
  """

  if color_string.startswith('%') and len(color_string) == 4:
    try:
      r, g, b = map(int, color_string[1:])
    except ValueError:
      pass
    else:
      if r < 6 and g < 6 and b < 6:
        return LutColor.from_rgb(r, g, b)

  elif color_string.startswith('$') and len(color_string) <= 4:
    try:
      index = int(color_string[1:])
    except ValueError:
      pass
    else:
      if index >= 0 and index < 256:
        return LutColor(index)

  elif color_string.startswith('#') and len(color_string) in (4, 7):
    parts = re.findall('.' if len(color_string) == 4 else '..', color_string[1:])
    if len(color_string) == 4:
      parts = [x*2 for x in parts]
    try:
      parts = [int(x, 16) for x in parts]
    except ValueError:
      pass
    else:
      return TrueColor(*parts)

  else:
    color_string = color_string.upper()
    bright = color_string.startswith('BRIGHT_') or color_string.startswith('BRIGHT ')
    if bright:
      color_string = color_string[7:]
    if hasattr(SgrColorName, color_string):
      return SgrColor(SgrColorName[color_string], bright)

  raise ValueError('unrecognizable color string: {!r}'.format(color_string))



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from typing import TextIO
import codecs
import io
import sys


class EncodedStreamReader(io.BufferedIOBase):
  """
  A readable file-like object that wraps a #TextIO object and converts it to a binary stream encoded with the specified
  *encoding* (or the default system encoding).

  Example:

  ```py
  import io
  from nr.utils.io.readers import EncodedStreamReader
  fp = EncodedStreamReader(io.StringIO('äöü'), 'utf-8')
  assert fp.read(1) == b'\xc3'
  assert fp.read(1) == b'\xa4'
  ```
  """

  def __init__(self, stream: TextIO, encoding: str = None, errors: str = 'strict'):
    encoding = encoding or getattr(stream, 'encoding', None) or sys.getdefaultencoding()
    self._stream = stream
    self._encoder = codecs.getencoder(encoding)
    self._errors = errors
    self._buffer = b''
    self.encoding = encoding

  def fileno(self):
    return self._stream.fileno()

  def read(self, n: int = None) -> bytes:
    if self._buffer and n is None:
      result, self._buffer = self._buffer, b''
    elif self._buffer:
      assert n is not None
      result, self._buffer = self._buffer[:n], self._buffer[n:]
      n -= len(result)
    else:
      result = b''
    if n is None or n >= 0:
      assert not self._buffer
      data = self._encoder(self._stream.read(-1 if n is None else n), self._errors)[0]
      if n is not None and len(data) > n:
        data, self._buffer = data[:n], data[n:]
      result += data
    return result




from __future__ import annotations

import logging


class SimpleFilter(logging.Filter):

  def __init__(self, name: str = '', contains: str | None = None, not_contains: str | None = None) -> None:
    super().__init__()
    self.contains = contains
    self.not_contains = not_contains

  def filter(self, record: logging.LogRecord) -> bool:
    if not super().filter(record):
      return False
    if self.contains and self.contains not in record.msg:
      return False
    if self.not_contains and self.not_contains  in record.msg:
      return False
    return True




import datetime
import typing as t
from dataclasses import dataclass
from nr.util.re import MatchFullError, match_full

if t.TYPE_CHECKING:
  import dateutil.relativedelta

SECONDS_PER_DAY = 3600 * 24
DAYS_PER_MONTH = 30.4375
DAYS_PER_YEAR = 365.25


@dataclass
class duration:
  """
  Represents an ISO 8601 duration.
  """

  years: int = 0
  months: int = 0
  weeks: int = 0
  days: int = 0
  hours: int = 0
  minutes: int = 0
  seconds: int = 0

  _fields = ['years', 'months', 'weeks', 'days', 'hours', 'minutes', 'seconds']

  def __post_init__(self) -> None:
    for k in self._fields:
      if getattr(self, k) < 0:
        raise ValueError('{} cannot be negative'.format(k))

  def __str__(self) -> str:
    parts = ['P']
    for value, char in [(self.years, 'Y'), (self.months, 'M'), (self.weeks, 'W'), (self.days, 'D')]:
      if value != 0:
        parts.append('{}{}'.format(value, char))
    has_t = False
    for value, char in [(self.hours, 'H'), (self.minutes, 'M'), (self.seconds, 'S')]:
      if value != 0:
        if not has_t:
          parts.append('T')
          has_t = True
        parts.append('{}{}'.format(value, char))
    return ''.join(parts)

  def total_seconds(self) -> float:
    """
    Computes the total number of seconds in this duration.
    """

    return (
      self.years * DAYS_PER_YEAR * SECONDS_PER_DAY +
      self.months * DAYS_PER_MONTH * SECONDS_PER_DAY +
      self.weeks * 7 * SECONDS_PER_DAY +
      self.days * SECONDS_PER_DAY +
      self.hours * 3600 +
      self.minutes * 60 +
      self.seconds)

  def as_timedelta(self) -> datetime.timedelta:
    """
    Returns the seconds represented by this duration as a #datetime.timedelta object. The arguments
    and keyword arguments are forwarded to the #total_seconds() method.
    """

    return datetime.timedelta(seconds=self.total_seconds())

  def as_relativedelta(self) -> 'dateutil.relativedelta.relativedelta':
    """
    Converts the #duration object to a #dateutil.relativedelta.relativedelta object. Requires
    the `python-dateutil` module.
    """

    from dateutil.relativedelta import relativedelta
    return relativedelta(years=self.years, months=self.months, weeks=self.weeks, days=self.days,
                         hours=self.hours, minutes=self.minutes, seconds=self.seconds)

  @classmethod
  def parse(cls, s: str) -> 'duration':
    """
    Parses an ISO 8601 duration string into a #duration object.

    Thanks to https://stackoverflow.com/a/35936407.
    See also https://en.wikipedia.org/wiki/ISO_8601#Durations
    """

    parts = s.split('T')
    if not s or s[0] != 'P' or len(parts) > 2:
      raise ValueError('Not an ISO 8601 duration string: {!r}'.format(s))

    part_one = parts[0][1:]
    part_two = parts[1] if len(parts) == 2 else ''

    fields = {}

    try:
      for number, unit in (x.groups() for x in match_full(r'(\d+)(D|W|M|Y)', part_one)):
        number = int(number)
        if unit == 'Y':
          fields['years'] = number
        elif unit == 'M':
          fields['months'] = number
        elif unit == 'W':
          fields['weeks'] = number
        elif unit == 'D':
          fields['days'] = number

      for number, unit in (x.groups() for x in match_full(r'(\d+)(S|H|M)', part_two)):
        number = int(number)
        if unit == 'H':
          fields['hours'] = number
        elif unit == 'M':
          fields['minutes'] = number
        elif unit == 'S':
          fields['seconds'] = number

    except MatchFullError:
      raise ValueError('Not an ISO 8601 duration string: {!r}'.format(s))

    return cls(**fields)




"""
Provides the #Scanner class which is convenient for scanning through items of a sequence; such as characters
in a text.
"""

from __future__ import annotations

import enum
import re
import typing as t
import typing_extensions as te


class Cursor(t.NamedTuple):
  offset: int
  line: int
  column: int

  def get_line_begin(self) -> Cursor:
    """ Returns a cursor pointing to the beginning of the current line. """
    return Cursor(self.offset - self.column + 1, self.line, 1)


class Seek(enum.Enum):
  SET = enum.auto()
  CUR = enum.auto()
  END = enum.auto()


class Scanner:
  """
  A convenient class for scanning through items of a sequence; such as characters in a text.
  """

  def __init__(self, text: str) -> None:
    self.text = text
    self._index = 0
    self._lineno = 1
    self._colno = 1

  def __repr__(self) -> str:
    return f'<Scanner at {self._lineno}:{self._colno}>'

  def __bool__(self) -> bool:
    return self._index < len(self.text)

  def __setattr__(self, key: str, value: t.Any) -> None:
    if key == '_index':
      min_value = 0
    elif key in ('_lineno', '_colno'):
      min_value = 1
    else:
      min_value = None
    if min_value is not None and value < min_value:
      raise RuntimeError(f'{key} cannot be set below {min_value}')
    object.__setattr__(self, key, value)

  @property
  def pos(self) -> Cursor:
    return Cursor(self._index, self._lineno, self._colno)

  @pos.setter
  def pos(self, cursor: Cursor) -> None:
    """ Moves the scanner back (or forward) to the specified cursor location. """

    if not isinstance(cursor, Cursor):
      raise TypeError(f'expected Cursor object {type(cursor).__name__}')
    self._index, self._lineno, self._colno = cursor

  @property
  def char(self) -> str:
    """ Returns the current character. Returns an empty string at the end of the text. """

    if self._index >= 0 and self._index < len(self.text):
      return self.text[self._index]
    else:
      return type(self.text)()

  def seek(self, offset: int, mode: te.Literal['set', 'cur', 'end'] | Seek = Seek.SET) -> None:
    """
    Moves the cursor of the Scanner to or by *offset* depending on the *mode*. The method is
    similar to a file's `seek()` method, but ensures that the line and column counts are tracked
    correctly.
    """

    if isinstance(mode, str):
      mode = Seek[mode.upper()]
    if mode not in Seek:
      raise ValueError(f'invalid mode: {mode!r}')

    # Translate the other modes into the 'set' mode.
    if mode == Seek.END:
      offset = len(self.text) + offset
    elif mode == Seek.CUR:
      offset = self._index + offset

    offset = max(0, min(len(self.text), offset))

    # Start counting from scratch.
    current_offset = 0
    colno = 1
    lineno = 1
    while current_offset != offset:
      newline_idx = self.text.find('\n', current_offset)
      if newline_idx >= offset or newline_idx < 0:
        left_newline_idx = self.text.rfind('\n', None, newline_idx if newline_idx >= 0 else len(self.text))
        colno = offset + 1 if left_newline_idx < 0 else offset - left_newline_idx
        current_offset = offset
        break
      else:
        current_offset = newline_idx + 1
        lineno += 1

    self._index, self._lineno, self._colno = current_offset, lineno, colno

  def next(self) -> str:
    """ Move on to the next character in the text. """

    if self._index >= len(self.text):
      return ''

    char = self.char
    if char == '\n':
      self._lineno += 1
      self._colno = 1
    else:
      self._colno += 1
    self._index += 1
    return self.char

  def readline(self) -> str:
    """ Reads a full line from the scanner and returns it. """

    start = end = self._index
    while end < len(self.text):
      if self.text[end] == '\n':
        end += 1
        break
      end += 1
    result = self.text[start:end]
    self._index = end
    if result.endswith('\n'):
      self._colno = 0
      self._lineno += 1
    else:
      self._colno += end - start
    return result

  def match(self, regex: t.Union[str, 're.Pattern'], flags: int = 0, *,
      _search: bool = False) -> t.Optional[t.Match[str]]:
    """
    Matches the *regex* from the current character of the *scanner* and returns the result. The
    scanners column and line numbers are updated respectively.
    """

    if isinstance(regex, str):
      regex = re.compile(regex, flags)
    match = (regex.search if _search else regex.match)(self.text, self._index)
    if not match:
      return None
    start, end = match.start(), match.end()
    if not _search:
      assert start == self._index, (start, self._index)
    else:
      start = self._index
    lines = self.text.count('\n', start, end)
    self._index = end
    if lines:
      self._colno = end - self.text.rfind('\n', start, end)
      self._lineno += lines
    else:
      self._colno += end - start
    return match

  def search(self, regex: t.Union[str, 're.Pattern'], flags: int = 0) -> t.Optional['re.Match']:
    """
    Performs a regex search from the current position of the scanner. Note that searching in the
    scanner will potentially have you skip characters without consuming them.
    """

    return self.match(regex, flags, _search=True)

  def getmatch(self, regex: t.Union[str, 're.Pattern'], group: t.Union[int,str] = 0,
      flags: int = 0) -> t.Optional[str]:
    """
    The same as #Scanner.match(), but returns the captured group rather than
    the regex match object, or None if the pattern didn't match.
    """

    match = self.match(regex, flags)
    if match:
      return match.group(group)
    return None

  def getline(self, cursor: Cursor) -> str:
    """ Returns the contents of the current line marked by the specified cursor location. """

    start = cursor.offset - cursor.column
    end = self.text.find('\n', start)
    if end < 0:
      end = len(self.text)
    return self.text[start:end]




import typing as t
from nr.util.digraph import DiGraph, E, K, N


def remove_with_predecessors(graph: DiGraph[K, N, E], nodes: t.Iterable[K]) -> None:
  """ Remove the given nodes from the graph, and their predecessors if they are not inputs to nodes that are kept on
  the graph. """

  untangle: set[K] = set()

  for node_id in nodes:
    untangle.update(graph.predecessors(node_id))
    del graph.nodes[node_id]

  while untangle:
    next_stage: set[K] = set()
    for node_id in untangle:
      if node_id not in graph.nodes: continue
      if not graph.successors(node_id):
        next_stage.update(graph.predecessors(node_id))
        del graph.nodes[node_id]
    untangle = next_stage




from __future__ import annotations

import errno
import grp
import os
import pwd
import signal
import sys
import time
import typing as t


def process_exists(pid: int) -> bool:
  """ Checks if the processed with the given *pid* exists. Returns #True if
  that is the case, #False otherwise. """

  if pid == 0:
    return False
  try:
    os.kill(pid, 0)
  except OSError as exc:
    if exc.errno == errno.ESRCH:
      return False
  return True


def process_terminate(pid: int, allow_kill: bool = True, timeout: int = 10) -> bool:
  """ Terminates the process with the given *pid*. First sends #signal.SIGINT,
  followed by #signal.SIGTERM after *timeout* seconds, followed by
  #signal.SIGKILL after *timeout* seconds if the process has not responded to
  the terminate signal.

  The fallback to kill can be disabled by setting *allow_kill* to False.
  Returns True if the process was successfully terminated or killed, or if
  the process did not exist in the first place. """

  def _wait(timeout):
    tstart = time.perf_counter()
    while (time.perf_counter() - tstart) < timeout:
      if not process_exists(pid):
        return True
      time.sleep(0.1)
    return False

  try:
    os.kill(pid, signal.SIGINT)
    if _wait(timeout):
      return True
    os.kill(pid, signal.SIGTERM)
    if _wait(timeout):
      return True
    if allow_kill:
      os.kill(pid, signal.SIGKILL)
      return _wait(timeout)
    return False
  except OSError as exc:
    if exc.errno == errno.ESRCH:
      return True
    raise


def getpwgrnam(user: str | None, group: str | None) -> tuple[str | None, int | None, int | None]:
  """ A combination of #pwd.getpwnam() and #pwd.getgrnam(), where *group*,
  if specified, overrides the group ID of the *user*. Returns a tuple of
  the user's home folder, the user ID and the group ID. """

  home, uid, gid = None, None, None
  if user:
    record = pwd.getpwnam(user)
    home, uid, gid = record.pw_dir, record.pw_uid, record.pw_gid
  if group:
    gid = grp.getgrnam(group).gr_gid
  return home, uid, gid


def replace_stdio(stdin: t.TextIO | None = None, stdout: t.TextIO | None = None, stderr: t.TextIO | None = None) -> None:
  """ Replaces the file handles of stdin/sdout/stderr, closing the original
  file descriptors if necessary. """

  if stdin:
    os.dup2(stdin.fileno(), sys.stdin.fileno())
  if stdout:
    os.dup2(stdout.fileno(), sys.stdout.fileno())
  if stderr:
    os.dup2(stderr.fileno(), sys.stderr.fileno())


def detach() -> None:
  """ Detaches the current process from the parent process. This function
  requires #os.setsid() and thus works only on Unix systems. """

  os.setsid()


def spawn_fork(func: t.Callable[[], t.Any], detach: bool = True) -> None:
  """ Spawns a single fork process and calls *func*. If *detach* is #True,
  the fork will be detached first (note that this process will still be killed
  by it's parent process if it doesn't exist gracefully).

  This is useful if *func* spawns another process, which will then behave like
  a daemon (as it will NOT be killed if the original process dies). """

  if not callable(func):
    raise TypeError('func is of type {} which is not callable'.format(
      type(func).__name__))

  pid = os.fork()
  if pid > 0:
    # Return to the original caller
    return
  if detach:
    os.setsid()
  func()
  os._exit(os.EX_OK)


def spawn_daemon(func: t.Callable[[], None]) -> None:
  """ Spawns a daemon process that runs *func*. This performs two forks to
  avoid the parent process killing the process that runs *func*.

  Note that this is only needed if you want to run a Python function as a
  daemon. If you were to spawn another process that should act as a daemon,
  you only need to fork once as the subprocess will then be the second "fork".
  """

  # TODO (@NiklasRosenstein): It would be great if the second fork could
  #   somehow report it's process ID to the original caller.

  if not callable(func):
    raise TypeError('func is of type {} which is not callable'.format(
      type(func).__name__))

  pid = os.fork()
  if pid > 0:
    # Return to the original caller.
    return
  os.setsid()
  pid = os.fork()
  if pid > 0:
    # Exit from second parent
    os._exit(os.EX_OK)
  func()
  os._exit(os.EX_OK)
